
# Wafl Test Library

## Concepts

Wafl tests are written as special Wafl programs, which may run all alone, but to have a well formatted result, it is expected to run them by a *Test Runner*. Each test program should contain a *Test Suite*. A number of test suites may be collected in a directory representing a *Test Collection*. A test collection may contain sub-folders (sub-collections).

Test Runners preprocess each suite before the evaluation. Thus, the suites have to follow some rules.

## Test Suite

A "test suite" is a program defined as a list of sections:

    Suite = List[Section]

A section is a record containing a section name and a list of tests:

    Section = { sectionName: <SECTION_NAME>, tests: List[Test] }

Each test must be a call to a test function, where a test function name must begin with `test` (case sensitive) and may belong to a library. A test function has to evaluate a record:

    { ok: Bool, desc: String }

An example of a full test suite is:

    [{
        sectionName: "One", 
        tests: [
            test( 3>2 ),
            test( 23 > 1 ),
            test( abs(2) == 2 ),
            testEq( 23, 1 ),
            testEq( "2" + "3", "23" )
        ]
    },{ 
        sectionName: "Two", 
        tests: [
            lib1::test( 3>2 ),
            testMyTst( 23 > 1 ),
            test( abs(2) == 2 ),
            testEq( 23, 1 ),
            testEq( "2" + "3", "23" )
        ]
    }]

If a test program consists only of a test suite, it will run
like any other program, but the test report will not be 
a nice looking one, and thus not fully usable.

## Preprocessing

Each test suite is preprocessed before the evaluation. All test functions are replaced by a richer construction to provide for a better reporting.

### IMPORTANT!

Because of the preprocessor limitations, the two important restrictions are imposed:

* Subdefinitions and where-expressions are not allowed inside 
of the test suite. If subdefinitions are required,
please define all names *after* the test suite:

        [{...test suite...}]
        where{
            ...
        }

* Comments must not include parenthesis. In some specific cases, especially after the last test in a section, a comment containing a closing parenthesis may induce the faulty preprocessing.

## Test Functions

Wafl library `test.wlib` defines some simple test functions:

* `test( exp )` - checks if the `exp` value is *true*;
* `testFalse( exp )` - checks if the `exp` value is *false*;
* `testEq( exp1, exp2 )` - checks if the values of the given expressions are equal;
* `testEqEps( exp1, exp2, eps )` - checks if the floating point values of `exp1` and `exp2` are in the `eps` range;
* `testDiff( exp1, exp2 )` - checks if the values of the given expressions are different.
* `testLT( exp1, exp2 )` - checks if the values of `exp1` is less than the value of `exp2`;
* `testLE( exp1, exp2 )` - checks if the values of `exp1` is less than or equal to the value of `exp2`;
* `testGT( exp1, exp2 )` - checks if the values of `exp1` is greater than the value of `exp2`;
* `testGE( exp1, exp2 )` - checks if the values of `exp1` is greater than or equal to the value of `exp2`;

It is simple to add new test functions. A user generated test function must complain to the following rules:

* Test function has to return a record:

        { ok: <boolean result>, desc: <string description> }

* The boolean element `ok` is the test result. It's value is `true` if test passes.

* The string element `description` is a textual description of the test. It should clearly describe why the test fails.

* The arguments of a test function are already evaluated, so the test has to do the simple comparison and create the description.

* Test function name must begin with `test`. However, it may be defined in a library.

For example, the `testDiff` test is implemented as follows:

    testDiff( arg1, arg2 ) = {
        ok: arg1 != arg2,
        desc: arg1.asString() + " != " + arg2.asString()
    };

For example, the following test function may be defined in the `where` section of the test program:

    testStrBeginsWith( s1, s2 ) = {
        ok: s1.strPos(s2) == 0,
        desc: '"' + s1 + '" does not begin with '" + s2 +'"'
    };

## Running a Test Suite

To run a standalone test program, it is enough to define a suite.
However, the suite value is not very readable. The better choice is
to use a *test runner*.

However, there is a universal function `wt::reportResults()` that will return a suite for a standalone version of the tests (i.e. it bahaves as an identity function), but will be automatically replaced with the appropriate one and process the results if a runner is used.

    [{...test suite...}]
    ->wt::reportResults()
    where {
        wt = library file 'Test.wlib';
    }

## Using a Test Runner

Test runner is a program that runs one or many test suites and generates a report. A simple test runner can be defined as follows:

    wtr::runAllTestsInSubfolders( '.', "waflTestReport.txt" )
    where {
        wtr = library file 'lib/testRunner.wlib';
    }

Function `wtr::runAllTestsInSubfolders( dir, repfile )` runs all tests in a collection and generates a report:

* `dir` is the base directory for the test collection. All `*.wafl` files in this directory and its direct subdirectories will be processed.
* `repfile` is report file name.

All test programs files will run in directory where test runner is started. Please define test programs to use relative paths to libraries.

## Default Test Runner

The easiest way to invoke the test runner is to use the default test runner:

    wtr::defaultRunner()
    where {
        wtr = library file 'TestRunner.wlib';
    }

The default test runner reads the command line arguments and selects the files to process:

    <runnerProgram> <target> <options>

    <target> := [<dirname> | <filename>]
    <option> := [ -detailed | -full | -lastreport | -repfile:<filename> ]

1. If a directory is specified as a target, the runner will process all Wafl test programs in the given directory and its direct subdirectories.
2. Default target is `./testCollection` directory.
3. If a filename is specified as a target, the runner will process only the given test file.
4. `-repfile:<filename>` options specifies that the report will be writen in the `<filename>`. The default report filename is `./waflTestReport.txt`.
5. `-detailed` options specifies that after the test running the final detailed report will be generated. Only the failed tests will be listed.
6. `-full` options specifies that after the test running the final detailed report will be generated. All tests, including both the passed and failed tests will be listed.
7. `-lastreport` option specifies that no test will run. Only the last report will be printed.
8. `-help` option lists the runner options.
